#ifndef __QOOB_H__
#define __QOOB_H__
#include <math.h>
#include <stdlib.h>
#include <windows.h>

// ********************************************************//
// COMMENT THIS LINE TO USE FULL LIBRARY (eg in a modeller)
// WHEN MAKING AN INTRO UNCOMMENT THE LINE

//#define INTROMODE

// You can of course set this in your compiler
// ********************************************************//

#define MAXRANGE 1e6
#define PI 3.14159265
#define DTR PI/180.0f
#define MN(a,b) (((a)<(b))?(a):(b))
#define MX(a,b) (((a)>(b))?(a):(b))
#define LRP(a,b,t) ((1.0f-(t))*(a)+(t)*(b))
#define BTW(a,b,c) ( ( ((c)>=(a)) && ((c)<=(b)) ) ? 1:0 )
#define SMOOTHSTEP(x) ((x) * (x) * (3 - 2 * (x)))
#define RCLAMPF(v) ( (v)<(-8.0f)? -8.0f : (v)>(8.0f) ? (8.0f) : (v) )
// The maximum size of meshes supported by qoob library...its a size coding thing.
#define MAXSIZE 250000

// a qoob object vert short form macro : object,quad,vertex(0..3)
#define QV(o,qi,vi) ((o)->v[(o)->q[(qi)][(vi)]])
// a qoob objevt vertex normal short form macro: object,quad,vertex(0..3)
#define QN(o,qi,vi) ((o)->n[(o)->q[(qi)][(vi)]])

typedef unsigned int uint;
typedef unsigned char uchar;

// vertex used in qoob
typedef struct vert_s { float x,y,z;} Vert;
// quad used in qoob.. all qoob meshes are composed only of quads
typedef uint Quad[4];
// a qoob mesh - the unpacked/uncompressed model that can be rendered
typedef struct obj_s {
   uint  nv;    // number of vertices in object
   uint  nq;    // number of quads in model
   Quad  q[MAXSIZE];    // the quad indices a,b,c,d (nq*4)
   Vert  v[MAXSIZE];    // the vertices x,y,z (nv*3)
   Vert  n[MAXSIZE];    // the vertex normals x,y,z (nv*3)
   Vert  fn[MAXSIZE];   // the face normals x,y,z (nq*3)
   uint  sel[MAXSIZE];  // which polygons are selected
} Object;

// The maximum number of qoob modelling commands used for qoob compressed models
#define MAXCOMMANDS 256

#pragma pack(push) // thanks to Ville Krumlinde for byte alignment advice
#pragma pack(1)
// a qoob compressed/packed model format ... 
// this struct is for storing a series of qoob lib modelling commands
typedef struct qoobo_s{
	uchar ncom;					// numnber of commands
	uchar com[MAXCOMMANDS];		// the commands
	uchar data[MAXCOMMANDS];	// data for commands that take a single float but comrepssed to char size
	uint picks[MAXCOMMANDS];	// picking data
} qoobObj;
#pragma pack(pop)

/////////////// HELPER FUNCTIONS ////////////////////////////
// vector
extern void add(Vert *,Vert *, Vert *);				//v3 = v2+v1
extern void sub(Vert *,Vert *, Vert *);				//v3 = v2-v1
extern void mul(Vert *,Vert *, Vert *);				//v3 = v2*v1
extern void smadd(Vert *,float ,Vert *, Vert *);	//v3 = v1*scalar + v2
extern void set(Vert *,float);						//all elements of v3 = scalar
extern void set3(Vert *,float, float, float);		//each element of v3 is set to each scalar
extern void copy(Vert *, Vert *);					//v2 = v1
extern void lerp(Vert *, Vert *, float, Vert *);	//v3 = linear interp (v1,v2, scalar)
extern void smul(Vert *, float);					//v1 = v1*scalar
extern float dot(Vert *, Vert *);					//returns v1.v2
//unusual! 
//cross ADDS the cp of v1,v2 into v3..zero v3 if u need absolute
extern void cross(Vert *, Vert *, Vert *);			//v3 += v1 x v2   !!!
extern float len2(Vert *);							//return lenght squared of v1
extern float len(Vert *);							//return length of v1
extern void norm(Vert *);							//v1 = ||v1||
// other
extern int  selected	(Object *,uint);			// is a particluar quad selected?uint is quad id
extern void setnormals	(Object *);					// create vertex and face normals - NOT normalised
extern void centroid (Object *, uint, Vert *);		// vertex = centre of quad (uint) of model
extern void msys_memset(void *, int , int);			// from iq, set memory values to val (first int param) amount time (second)
extern void msys_memcpy(void *, const void *, int); // from iq, copy memory (dest, src, amount) 

///////////////////// MODELLING FUNCTIONS /////////////////////
// All fucntions below have floats even though most are ignored
// this helps simplify code and make things smaller :-)

// selection handling for modelling
extern void selectAll		(Object *, float);	// select all quads in model, float is ignored
extern void deselectAll		(Object *, float);	// deselect all quads, float is ignored
extern void invselectAll	(Object *, float);	// invert selection state of all quads, float is ignpred
extern void growselect		(Object *, float);	// add quads to selection which are next to ones already selected, float ignored
extern void adjselect		(Object *, float);	// deselect current quads and select thos that were next to them, float ognored
extern void xselect			(Object *, float);	// add quads which are have similar X values to LAST CHOSEN QUAD, float ignored
extern void yselect			(Object *, float);  // add quads which are have similar Y values to LAST CHOSEN QUAD, float ignored
extern void zselect			(Object *, float);  // add quads which are have similar Z values to LAST CHOSEN QUAD, float ignored
extern void invselectQuad	(Object *, uint);	// if quad (uint) is selected, it becomes deselected or vice versa

// WARNING: in the following functions, all float parameters MUST BE clamped to -8.0 to 8.0
// ..other values will have unexpected results
//translation functions
extern void transQuadsN (Object *, float);	//move all selected quads along their normals by amount float
extern void transObjX	(Object *, float);	//move all quads along X axis by amount float
extern void transObjY	(Object *, float);	//translate all quads along Y axis by amount float
extern void transObjZ	(Object *, float);	//translate all quads along Z axis by amount float
extern void transQuadsX (Object *, float);	//move all selected quads along X axis by amount float
extern void transQuadsY (Object *, float);  //translate all selected quads along Y axis by amount float
extern void transQuadsZ (Object *, float);	//translate all selected quads along Z axis by amount float
//scales fucntions
extern void scaleObjX	(Object *, float);	//scale entire object in X axis by float amount (1.0=no scale)
extern void scaleObjY	(Object *, float);  //scale entire object in Y axis by float amount (1.0=no scale)
extern void scaleObjZ	(Object *, float);  //scale entire object in Z axis by float amount (1.0=no scale)
extern void scaleObjUniform(Object *, float ); //scale entire object by float amount (1.0=no scale)
extern void scaleQuadsX	(Object *, float);	//scale selected quads in X axis by float amount (1.0=no scale)
extern void scaleQuadsY	(Object *, float);	//scale selected quads in Y axis by float amount (1.0=no scale)
extern void scaleQuadsZ	(Object *, float);	//scale selected quads in Z axis by float amount (1.0=no scale)
extern void scaleQuadsUniform(Object *, float); //scale selected quads by float amount (1.0=no scale)
extern void scaleQuadsCen (Object *, float); //very dangerous - not for general use...
//rotate functions
extern void rotQuadsX	(Object *, float); //rotate selected quads about X axis by ((float)/16.0f*360.0f)*3.14159f/180.0f;
extern void rotQuadsY	(Object *, float); //rotate selected quads about Y axis by ((float)/16.0f*360.0f)*3.14159f/180.0f;
extern void rotQuadsZ	(Object *, float); //rotate selected quads about Z axis by ((float)/16.0f*360.0f)*3.14159f/180.0f;
//other
extern void subdObj		(Object *, float); // subdivide every quad into four equal ones, no displacement, float ignored
										   // if original quad was selected , four new qauds will be and vice versa
extern void smoothObj	(Object *, float); // catmull clark subdivision - handles holes to some extent, float ignored
										   // if original quad was selected , four new qauds will be and vice versa
extern void addQuads	(Object *, float); // insert a selected quad in the centre of all selected quads, float ignored
extern void delQuads	(Object *, float); // delete selected quads, float ignored
extern void addcubeObj	(Object *, float); // adds a qoob from -1.0 to 1.0. It is selected when added.

extern void unpackObj	(qoobObj *, Object *);	// decompress a qoob object into a mesh
extern void replay		(Object *);				// plays the qoob lib internal object into a mesh (provided)
extern uint	adjacentQuad(Object *, uint, uint); // are the two uint quad ids next to each other in the mesh

#ifndef INTROMODE
// in here are routines available when not in an intro, eg for qoob modeller
extern void recordon	(void);		// switch on recording of commands
extern void recordoff	(void);     // and off
extern void rewind		(void);		// deletes last recorded action
extern void initrecord  (void);		// start recording fresh, deletes all recorded actions.. 
									// should be called before first call to recordon 
extern void exportObj	(char *);	// save recorded object in C form to file qoobject.h, the string is the name of the object
extern void loadObj		(LPCTSTR,Object *);	// load a qoob compressed object from file. First param is file name
extern void saveObj		(LPCTSTR);	// save a compressed qoob object, parameter is file name

extern int  numberQoobCommands(void);	// how many commands has qoob lib recorded?
extern int  numberPicks(void);			// how many picks?
extern int  numberSelected(Object *);	// how many quads are selected in the given mesh

#endif //INTROMODE

#endif // __QOOB_H__

